// (c) Copyright 2012 Cloudera, Inc. All rights reserved.
define([
  "underscore",
  "cloudera/Util",
  "cloudera/cmf/charts/TimeSeriesViewModel",
  "cloudera/cmf/charts/Facets",
  "cloudera/cmf/charts/Plot",
  "cloudera/layout/Resizer",
  "cloudera/common/I18n",
  "cloudera/common/charts/rickshaw/FormatUtils",
  "cloudera/common/charts/rickshaw/Visualizer",
  "cloudera/common/Monitor",
  "cloudera/common/charts/include/TimeSeriesUtil",
  "cloudera/common/charts/include/TimeSeriesLinearSampler"
], function(_, Util, TimeSeriesViewModel, Facets, Plot, Resizer, I18n, FormatUtils, Visualizer, Monitor, TimeSeriesUtil, TimeSeriesLinearSampler) {

  // If we are trying to show more streams than this we should not show
  // the legend for this chart.
  var SHOW_LEGEND_NUMBER_OF_STREAMS_THRESHOLD = 50;

  /**
   * Renders the response of a single tsquery into a group of charts
   * using the rickshaw's JavaScript library.
   *
   * This creates n rickshaw specific visualizer objects.
   * In the future, this will also be the place to render morris/Raphael charts
   * in IE8.
   *
   * options = {
   *   plot:         (required) an object that is deserialized from the JSON representation of a Plot object.
   *   container:    (required) "selector or element of the container DOM"
   *   context:        (optional) a mapping of placeholder variable names to values.
   *   resizeContainer:(optional) a resize element
   *   resizeLabel:    (optional) a element that indicates the width after resize.
   *   enableFacets:   (optional) true|false, whether facets should be enabled, default false.
   *   enableEditing:  (optional) true|false, whether the charts can be edited, default false.
   *   enableCloning:  (optional) true|false, whether the charts can be cloned, default false.
   *   enableRemoving: (optional) true|false, whether the charts can be removed, default false.
   *   enableEnlarging:(optional) true|false, whether the chart can be enlarged, default true.
   * };
   */
  function TimeSeriesVisualizer(options) {
    var self = this, $container = $(options.container);

    var $chartsContainer = $container.find(".charts-container");
    if ($chartsContainer.length === 0) {
      $chartsContainer = $('<div class="charts-container"></div>').appendTo($container);
    }
    var $facetsContainer = $container.find(".facets-container");
    if (options.enableFacets && $facetsContainer.length === 0) {
      $facetsContainer = $('<div class="facets-container"></div').appendTo($container);
    }

    if (options.enableFacets) {
      self.facets = new Facets({
        container: $facetsContainer,
        plot: options.plot
      });
    }

    self.viewModel = new TimeSeriesViewModel(options.plot);
    self.viewModel.setVisualizer(self);
    self.viewModel.setContext(options.context);

    self.getViewModel = function() {
      return self.viewModel;
    };

    if (options.resizeContainer) {
      var resizeCallback = function(scale) {
        var width = scale,
          height = self.viewModel.height() * scale / self.viewModel.width();
        self.viewModel.setDimension(width, height);
      };

      self.resizer = new Resizer({
        element: options.resizeContainer,
        min: 50,
        max: 1000,
        step: 50,
        scale: self.viewModel.width(),
        callback: _.debounce(resizeCallback, 500)
      });
    }


    /**
     * Returns true if there are charts.
     */
    self.hasCharts = function() {
      return self.vizs && self.vizs.length > 0;
    };

    self.onEditPlot = function(evt) {
      $.publish("editPlot", [self.getViewModel().getPlot(), $chartsContainer]);
      evt.preventDefault();
      evt.stopPropagation();
    };

    self.onClonePlot = function(evt) {
      $.publish("clonePlot", [self.getViewModel().getPlot(), $chartsContainer]);
      evt.preventDefault();
      evt.stopPropagation();
    };

    self.onRemovePlot = function(evt) {
      var callback = function() {
        $chartsContainer.fadeOut('slow', function() {
          $chartsContainer.remove();
        });
      };
      $.publish("removePlot", [self.getViewModel().getPlot(), $chartsContainer, callback]);
      evt.preventDefault();
      evt.stopPropagation();
    };

    self.onCloneChart = function(evt, groupTs) {
      $.publish("clonePlot", [self.getViewModel().getPlot(), $chartsContainer]);
      evt.preventDefault();
      evt.stopPropagation();
    };

    self.getChartCount = function(groupedTimeSeries) {
      var count = 0;
      _.each(groupedTimeSeries, function(v, k) {
        count += 1;
      });
      return count;
    };

    // Given a jQuery-wrapped chart container, add the HTML nodes to it
    // necessary to hold a legend.
    self.addLegendContainer = function($chartContainer) {
      var $legendContainer = $('<div class="legend"/>');
      if (self.viewModel.showYAxis()) {
        $legendContainer.css('padding-left', TimeSeriesVisualizer.Y_AXIS_WIDTH + "px");
      }
      $chartContainer.append($legendContainer);
    };

    // We might not show a legend if there are too many streams, despite what
    // the options say.
    var willShowLegend = function(groupTs) {
      return self.viewModel.showLegend()
        && groupTs.length <= SHOW_LEGEND_NUMBER_OF_STREAMS_THRESHOLD;
    };

    /**
     * Renders a time series response object, which has 0 or more time series.
     */
    self.render = function(timeSeriesResponse) {
      var monitor = new Monitor("TimeSeriesVisualizer.render");
      // Collapse graph data by facet
      var facetting = self.viewModel.facetting();

      // Remove all the empty timeseries,
      // and store a reference for facetting purposes.
      monitor.open("filtering empty ts");
      var nonEmptyTimeSeriesArray = _.filter(timeSeriesResponse.timeSeries, function(ts) {
        return ts.data && ts.data.length > 0;
      });
      monitor.close("filtering empty ts");

      monitor.open("regrouping ts");
      // a map of groupKey -> [ {
      //   data: [{x: .., y: ...}],
      //   metadata: {}
      // } ].
      var groupedTimeSeries = self.regroupTimeSeries(nonEmptyTimeSeriesArray, facetting);
      monitor.close("regrouping ts");

      monitor.open("setting up page");
      // Clear the page.
      $chartsContainer.empty();
      // Only append to this fragment in this method. We don't want to cause
      // page reflows.
      var fragment = document.createDocumentFragment();
      self.vizs = [];

      var chartCount = self.getChartCount(groupedTimeSeries);
      var mainTitle = self.viewModel.title();
      var titleContainerElement = self.createTitleContainer(
        mainTitle, 'h2',
        options.enableEditing,
        options.enableCloning && (chartCount > 1 && facetting === Plot.FACETTING_NONE),
        options.enableRemoving);
      fragment.appendChild(titleContainerElement);
      var $titleContainer = $(titleContainerElement);
      $titleContainer.find(".edit-chart").click(self.onEditPlot);
      $titleContainer.find(".clone-chart").click(self.onClonePlot);
      $titleContainer.find(".remove-chart").click(self.onRemovePlot);

      var firstTitle = true;
      monitor.close("setting up page");
      monitor.open("rendering charts");
      // For each group of time series, render a single chart.
      $.each(groupedTimeSeries, function(groupKey, groupTs) {
        monitor.open("preparing " + groupKey);
        var chartContainerElement = document.createElement('div');
        chartContainerElement.className = 'chart-container';
        var $chartContainer = $(chartContainerElement);
        var title = self.getTitle(facetting, groupKey, groupTs);

        // If there is more than one stream in the group we need to
        // sample both streams onto the same x (this is a requirement of rickshaw
        // which is build on top of d3 stackLayout). If there is only one stream
        // then we just downsample it to width of the chart, if necessary.
        monitor.open("sampling");
        TimeSeriesLinearSampler.sampleTimeSeriesGroup(groupTs,
                                                      self.viewModel.width());
        monitor.close("sampling");

        var yAxisLabelElement = self.createYAxisLabel(groupTs);
        chartContainerElement.appendChild(yAxisLabelElement);

        // For individual charts, we cannot edit/remove.
        var titleContainerElement = self.createTitleContainer(
          mainTitle === title ? "" : title, 'h3', false,
          options.enableCloning, false);
        var $titleContainer = $(titleContainerElement);
        $titleContainer.find(".clone-chart").click(function(evt) {
          self.onCloneChart(evt, groupTs);
        });

        firstTitle = false;

        var chartElement = document.createElement('div');
        chartElement.className = 'chart';
        var $chart = $(chartElement);
        if (self.viewModel.showYAxis()) {
          var yAxisElement = document.createElement('div');
          yAxisElement.className = 'yAxis';
          var $yAxis = $(yAxisElement);
          $chart.css("left", TimeSeriesVisualizer.Y_AXIS_WIDTH + "px");
          $yAxis.css("width", TimeSeriesVisualizer.Y_AXIS_WIDTH + "px");
          $yAxis.css("height", self.viewModel.height() + "px");
          $chartContainer.append($yAxis);
        }
        if (_.isUndefined(options.enableEnlarging) || options.enableEnlarging) {
          $chartContainer.click(function(evt) {
            $.publish("showPlotDetails", [self.getViewModel().getPlot(), groupTs, title]);
          });
        }

        chartContainerElement.appendChild(chartElement);
        if (willShowLegend(groupTs)) {
          self.addLegendContainer($chartContainer);
        }
        chartContainerElement.appendChild($titleContainer[0]);

        var renderOptions = {
          container: $chartContainer,
          width: self.viewModel.width(),
          height: self.viewModel.height(),
          yAxisWidth: TimeSeriesVisualizer.Y_AXIS_WIDTH,
          ymin: self.viewModel.ymin(),
          ymax: self.viewModel.ymax(),
          interpolation: self.viewModel.interpolation(),
          offset: self.viewModel.offset(),
          renderer: self.viewModel.renderer(),
          showYAxis: self.viewModel.showYAxis(),
          showHoverDetail: self.viewModel.showHoverDetail(),
          showLegend: willShowLegend(groupTs)
        };
        fragment.appendChild(chartContainerElement);
        monitor.close("preparing " + groupKey);
        monitor.open("rendering " + groupKey);
        try {
          monitor.open("creating visualizer for " + groupKey);
          var viz = self.createVisualizer(renderOptions);
          monitor.close("creating visualizer for " + groupKey);
          monitor.open("visualizer.render for " + groupKey);
          viz.render(groupTs, monitor);
          self.vizs.push(viz);
          monitor.close("visualizer.render for " + groupKey);
        } catch (ex) {
          console.log(ex);
        }
        monitor.close("rendering " + groupKey);

      });
      monitor.close("rendering charts");
      monitor.open("facets");
      self.adjustWidth(self.viewModel.width(), fragment);

      if (self.facets) {
        self.timeSeriesResponse = timeSeriesResponse;
        self.facets.update(nonEmptyTimeSeriesArray);
      }
      monitor.close("facets");
      monitor.log();
      $chartsContainer[0].appendChild(fragment);
    };

    /**
     * When there is no facetting, use the label
     * of the time series.
     * When plotting everything in one chart, use the plot title.
     * Otherwise, use the explicit facetting attribute.
     * Otherwise, we use the title of the plot.
     */
    self.getTitle = function(facetting, groupKey, groupTs) {
      if (facetting === Plot.FACETTING_NONE) {
        var ts = _.first(groupTs);
        if (ts && ts.metadata) {
          return ts.metadata.label || "";
        } else {
          return "";
        }
      } else if (facetting === Plot.FACETTING_SINGLE_PLOT) {
        // Use the title of the plot.
        return self.viewModel.title();
      } else {
        // Use the explicit groupKey attribute.
        return groupKey || I18n.t("ui.na");
      }
    };

    self.createMenuEntry = function(clazz, icon, text) {
      var liElement = document.createElement('li');
      var aElement = document.createElement('a');
      aElement.className = clazz;
      aElement.href = '#';
      var iconElement = document.createElement('i');
      iconElement.className = icon;
      aElement.appendChild(iconElement);
      var spanElement1 = document.createElement('span');
      spanElement1.innerHTML = " " + text;
      aElement.appendChild(spanElement1);
      liElement.appendChild(aElement);
      return liElement;
    };

    self.createYAxisLabel = function(groupTs) {
      var yAxisLabel = FormatUtils.getYAxisLabel(groupTs);
      var yAxisLabelElement = document.createElement("div");
      yAxisLabelElement.innerHTML = yAxisLabel;
      yAxisLabelElement.className = "yAxisLabel";
      yAxisLabelElement.title = yAxisLabel;
      return yAxisLabelElement;
    };

    self.createTitleContainer = function(title, tag, enableEditing, enableCloning, enableRemoving) {
      var titleContainerElement = document.createElement('div');
      var titleElement = document.createElement(tag);
      titleElement.className = 'chart-title';
      titleElement.innerHTML = title || '&nbsp;';
      titleElement.title = title;

      if (enableEditing || enableCloning || enableRemoving) {
        var entries = 0;
        if (enableEditing) {
          entries++;
        }
        if (enableCloning) {
          entries++;
        }
        if (enableRemoving) {
          entries++;
        }
        var toolbarContainerElement = document.createElement('div');
        toolbarContainerElement.className = 'chart-toolbar';
        var menuElement = document.createElement('ul');
        if (enableEditing) {
          menuElement.appendChild(self.createMenuEntry("edit-chart", "icon-edit", I18n.t("ui.chart.edit")));
        }

        if (enableCloning) {
          menuElement.appendChild(self.createMenuEntry("clone-chart", "icon-share", I18n.t("ui.chart.clone")));
        }

        if ((enableEditing || enableCloning) && enableRemoving) {
          var dividerElement = document.createElement('li');
          dividerElement.className = 'divider';
          menuElement.appendChild(dividerElement);
        }

        if (enableRemoving) {
          menuElement.appendChild(self.createMenuEntry("remove-chart", "icon-trash", I18n.t("ui.chart.remove")));
        }

        if (entries > 1) {
          var btnGroupElement = document.createElement('div');
          btnGroupElement.className = 'btn-group alignRight';
          var dropDownElement = document.createElement('a');
          dropDownElement.className = 'btn btn-mini btn-info dropdown-toggle';
          dropDownElement.setAttribute('data-toggle', 'dropdown');
          dropDownElement.href = '#';
          var caretElement = document.createElement('span');
          caretElement.className = 'caret';
          dropDownElement.appendChild(caretElement);
          menuElement.className += ' dropdown-menu';
          toolbarContainerElement.appendChild(btnGroupElement);
          btnGroupElement.appendChild(dropDownElement);
          btnGroupElement.appendChild(menuElement);
        } else {
          menuElement.className += ' unstyled';
          toolbarContainerElement.appendChild(menuElement);
        }
        titleContainerElement.appendChild(toolbarContainerElement);
      }

      titleContainerElement.appendChild(titleElement);
      return titleContainerElement;
    };

    self.adjustWidth = function(width, container) {
      // TimeSeriesViewModel calls this method without
      // the container argument.
      if (container === undefined) {
        container = $chartsContainer;
      }
      self.adjustTitleWidth(width, container);
      self.adjustChartWidth(width, container);
    };

    self.adjustTitleWidth = function(width, container) {
      var $chartTitle = $(container).children().find(".chart-title");
      if (width < 160) {
        $chartTitle.hide();
      } else {
        $chartTitle.css("width", (TimeSeriesVisualizer.Y_AXIS_WIDTH + width) + "px").show();
      }
    };

    self.adjustChartWidth = function(width, container) {
      var $chart = $(container).children().find(".chart");
      $chart.css("width", width + "px");
      // Update the resize label.
      $(options.resizeLabel).text("(" + width + ")");
    };

    /**
     * Expose this method for testing purposes.
     */
    self.createVisualizer = function(renderOptions) {
      return new Visualizer(renderOptions);
    };

    /**
     * Group the time series.
     */
    self.regroupTimeSeries = function(timeSeriesArray, facet) {
      var grouped = {};
      if (facet === Plot.FACETTING_SINGLE_PLOT) {
        grouped.all = timeSeriesArray;
      } else if (facet !== Plot.FACETTING_NONE) {
        var facets = facet.split(",");
        var extractKey = function(o) {
          var v = [];
          $.each(facets, function(_, g) {
            if (o.metadata.attributes[g]) {
              v.push(o.metadata.attributes[g]);
            } else if (o.metadata[g]) {
              v.push(o.metadata[g]);
            }
          });
          return v.join(",");
        };
        $.map(timeSeriesArray, function(ts, i) {
          var k = extractKey(ts);
          if(!grouped.hasOwnProperty(k)) {
            grouped[k] = [];
          }
          if (grouped[k]) {
            grouped[k].push(ts);
          }
        });
      } else {
        $.map(timeSeriesArray, function(ts, i) {
          grouped[i] = [ts];
        });
      }
      return grouped;
    };

    /**
     * Apply a new configuration to all the charts.
     */
    self.update = function(config) {
      if (self.vizs) {
        _.invoke(self.vizs, 'update', config);
      }
    };

    /**
     * Handles the facet clicking event.
     */
    var handle1 = $.subscribe("clickFacetGroup", function(facetting) {
      self.viewModel.facetting(facetting);
      self.render(self.timeSeriesResponse);
    });

    self.subscriptionHandles = [handle1];

    self.unsubscribe = function() {
      Util.unsubscribe(self);
    };
  }

  TimeSeriesVisualizer.Y_AXIS_WIDTH = 50;

  return TimeSeriesVisualizer;
});
